// SUBNETWORK HIERARCHIES OF BIOCHEMICAL NETWORKS ========================
// by Petter Holme, April 2002 ===========================================

#include "net.h"

extern NODE *node;
extern char fname[NAMESIZE];
extern PAR par;

// this struct contains vertex-specific information for the betweenness
// calculation . . .
typedef struct BN {
  int distance, np, *predecessor;
  double delta, through;
} BN;
  
//------------------------------------------------------------------------
// using Brandes' algorithm for calculating betweenness on directed graphs
// straightforwardly implemented from: Brandes.U. (2001) A faster
// algorithm for betweenness centrality. J. Math. Sociol., 25, 163-177.

void between (double *bnode) {
  int i, j, me, you, *start_check, *end_check, *check, *end_order, *order;
  BN *bn;

  // allocation . . .
  check = (int *) malloc(2 * par.n * sizeof(int));
  order = check + par.n;
  bn = (BN *) malloc(par.n * sizeof(BN));
  for (i = 0; i < par.n; i++) {
    bnode[i] = 0.0;
    if (node[i].indegree) bn[i].predecessor = (int *)
      malloc(node[i].indegree * sizeof(int));
  }

  // only considering paths between substances . . .
  for (i = 0; i < par.n - par.nreaction; i++) {
    // initialization . . .
    start_check = end_check = check;
    end_order = order;
    for (j = 0; j < par.n; j++) {
      bn[j].distance = -1;
      bn[j].np = 0;
      bn[j].delta = bn[j].through = 0.0;
    }
    bn[i].through = 1.0;
    bn[i].distance = 0;
    *end_check++ = i;

    // search shortest paths . . .
    do {
      me = *start_check++;
      *end_order++ = me;
      // following the paths in the outward direction . . .
      for (j = 0; j < node[me].outdegree; j++) {
        you = node[me].outneighbor[j];
        // 'you' found for the first time ? . . .
        if (bn[you].distance < 0) {
          *end_check++ = you;
          bn[you].distance = bn[me].distance + 1;
        }
        // shortest path to 'you' via 'me' ? . . .
        if (bn[you].distance == bn[me].distance + 1) {
          bn[you].through += bn[me].through;
          bn[you].predecessor[bn[you].np++] = me;
        }
      }
    } while (end_check > start_check);

    // collecting info for betweenness centrality . . .
    while (end_order > order) {
      me = *--end_order;
      for (j = 0; j < bn[me].np; j++) {
        you = bn[me].predecessor[j];
        bn[you].delta
          += ((1.0 + bn[me].delta) * bn[you].through) / bn[me].through;
      }
      if (i != me) bnode[me] += bn[me].delta;
    }
  }
  for (i = 0; i < par.n; i++) if (node[i].indegree) free(bn[i].predecessor);

  // rescaling for effective betweenness . . .
  for (i = par.n - par.nreaction; i < par.n; i++) {
    if (node[i].indegree) bnode[i] = bnode[i] / (double) node[i].indegree;
    else bnode[i] = 0.0; // with a complete database this 'else' isn't
    // needed (all reaction nodes has non-zero in-degree . . . )
  }
  free(bn); free(check);
}

//------------------------------------------------------------------------
// removing the reaction node with highest betweenness . . .
// (or rather deleting the links of the node with highest betweenness)

void attack (double *bnode) {
  int i, j, imax, you, me, istore = 0, iimx, *imx;
  double bmax = -1.0;

  for (i = 0; i < par.n; i++) bnode[i] = 0.0;
  
  between(bnode);
  
  imx = (int *) malloc(par.n * sizeof(int));

  // picking the reaction node of highest betweenness . . .
  for (i = 1; i <= par.nreaction; i++) {
    if (fabs(bmax - bnode[par.n - i]) < EPS) imx[istore++] = par.n - i;
    if (bmax < bnode[par.n - i]) {
      bmax = bnode[par.n - i]; imx[0] = par.n - i; istore = 1;}
  }

  // deconstruct (i.e. remove imx's links) . . .
  // . . . outward links
  for (iimx = 0; iimx < istore; iimx++) {
    imax = imx[iimx];
    for (i = 0; i < node[imax].outdegree; i++) {
      you = node[imax].outneighbor[i];
      me = 0;
      while (node[you].inneighbor[me] != imax) me++;
      for (j = me; j < node[you].indegree - 1; j++)
	node[you].inneighbor[j] = node[you].inneighbor[j + 1];
      node[you].indegree--;
    }
    if (node[imax].outdegree) free(node[imax].outneighbor);
    node[imax].outdegree = 0;
    
    // . . . inward links
    for (i = 0; i < node[imax].indegree; i++) {
      you = node[imax].inneighbor[i];
      me = 0;
      while (node[you].outneighbor[me] != imax) me++;
      for (j = me; j < node[you].outdegree - 1; j++)
	node[you].outneighbor[j] = node[you].outneighbor[j + 1];
      node[you].outdegree--;
    }
    if (node[imax].indegree) free(node[imax].inneighbor);
    node[imax].indegree = 0;
  }
}

//------------------------------------------------------------------------
