// SUBNETWORK HIERARCHIES OF BIOCHEMICAL NETWORKS ========================
// by Petter Holme, April 2002 ===========================================

// this program draws hierarchy trees in xfig's internal format
// see http://xfig.org it takes names of hi-datafiles as input
// and is conveniently called as: ls data/* | draw

#include "net.h"

#define LINE "2 1 0 3 32 7 50 0 -1 0.000 0 0 -1 0 0 2\n\t "
#define TAG "4 2 0 50 0 16 8 0.0000 4 120 90 -50 %d %d\\001\n"
#define TAG2 "4 2 33 50 0 18 8 0.0000 4 120 90 -50 %d %d\\001\n"
#define TEXTTAG "4 2 0 50 0 16 8 0.0000 4 120 90 -300 %d %s\\001\n"
#define DY 100 // sets the spacing in the y-direction . . .
#define RATIO 2.778 // sets the x to y ratio
#define NALLOC 10

// struct for current-cluster info . . .
// size -- the size of the cluster
// node -- an array of nodes belonging to the cluster
// tree -- the branching point above the cluster
typedef struct CL {
  int size, *node, tree;
} CL;

// struct for drawing the hiearchy-tree representing a branching-point
// in the hierarchy tree . . .
// level -- the height (x-coordinate) of branching point
// alloc -- the size allocated for 'kid'
//   kid -- an array of children to the branching point
//  nkid -- number of children
//     y -- the y-coordinate of the branching point
typedef struct TREE {
  int level, alloc, *kid, nkid, y;
} TREE;

NODE *node;
CL *cl;
int n, metamax, reactionmax, mode;

//------------------------------------------------------------------------
// parsing data files and constructing metabolic networks . . .

void read_net (char *modetag, char *orgtag) {
  int i, me, you, ti;
  char tname[15];
  FILE *fp;

  sprintf(tname, "%s/%s.dat", modetag, orgtag);
  fp = fopen(tname, "r");
  if (!fp) {
    fprintf(stderr, " error opening %s\n", tname);
    exit(-1);
  } else fprintf(stderr, " read %s\n", tname);

  // scanning the dimensions . . .
  // . . . number of nodes
  metamax = reactionmax = -1;
  while (fscanf(fp, "%d %d", &me, &you) == 2) {
    if (me < you) {
      if (you > reactionmax) reactionmax = you;
      if (me > metamax) metamax = me;
    } else {
      if (me > reactionmax) reactionmax = me;
      if (you > metamax) metamax = you;
    }
  }
  rewind(fp);
  reactionmax -= (M0 - 1);
  n = metamax + reactionmax;
  node = (NODE *) malloc(n * sizeof(NODE));
  for (i = 0; i < n; i++) node[i].indegree = node[i].outdegree = 0;

  // . . . degrees
  ti = M0 - metamax; // moving reaction nodes to their position in 'node'
  while (fscanf(fp, "%d %d", &me, &you) == 2) {
    if (me < you) {
      node[me - 1].outdegree++;
      node[you - ti].indegree++;
    } else {
      node[me - ti].outdegree++;
      node[you - 1].indegree++;
    }
  }
  rewind(fp);

  // allocating & re-zeroing . . .
  for (i = 0; i < n; i++) {
    if (node[i].outdegree) node[i].outneighbor =
	    (int *) malloc(node[i].outdegree * sizeof(int));
    if (node[i].indegree) node[i].inneighbor =
	    (int *) malloc(node[i].indegree * sizeof(int));
    node[i].indegree = node[i].outdegree = 0;
  }
  
  // constructing the network . . .
  while (fscanf(fp, "%d %d", &me, &you) == 2) {
    if (me < you) {
      node[me - 1].outneighbor[node[me - 1].outdegree++] = you - ti;
      node[you - ti].inneighbor[node[you - ti].indegree++] = me - 1;
    } else {
      node[me - ti].outneighbor[node[me - ti].outdegree++] = you - 1;
      node[you - 1].inneighbor[node[you - 1].indegree++] = me - ti;
    }
  }
  fclose(fp);
}

//------------------------------------------------------------------------

int main () {
  int i, itree, nclu, me, it, ivsize, event;
  int ilevel, itmp, jtmp, nctmp, xmax, ymax;
  int treealloc, j, k, l, oldilevel, dx, dy, ime, metree, jtree;
  int xt, found, wrote;
  static int *piv, *iv = (int *) NULL, *tmp, *ntmp, *ctmp;
  int *index;
  char fname[100], f2name[50], orgnames[200], orgtag[3], modetag[5];
  FILE *fp, *nfp;
  TREE *tree;

  modetag[0] = orgtag[0] = '\0';
  
  node = (NODE *) NULL;

  while (fgets(fname, 100, stdin) != NULL) {
    for (i = 0; i < 100; i++) if (fname[i] == '\n') {
      fname[i] = '\0';
      break;
    }

    // scanning the file name . . .
    // . . . determining organism name
    strncpy(orgtag, fname + 9, 2);
    if ((!isalpha(orgtag[0])) || (!isalpha(orgtag[1]))) {
      fprintf(stderr, " strange datafile '%s'\n", fname);
      exit(-1);
    }
    orgtag[2] = '\0';
    fp = fopen("orgnames", "r");
    i = 0;
    while (fgets(orgnames, 4, fp)) {
      i++;
      if (!(strncmp(orgnames, orgtag, 2))) {
	ime = i;
	break;
      }
    }
    fclose(fp);
    
    // . . . determining metabolic or whole-cell network
    if (!strncmp(fname + 5, "meta", 4)) {
      mode = META;
      strcpy(modetag, "meta");
    } else {
      mode = CELL;
      strcpy(modetag, "cell");
    }
   
    // scanning the dimensions . . .
    // . . . number of nodes & links
    metamax = reactionmax = -1;
    read_net(modetag, orgtag);

    // opening and reading the binary data file . . .
    fp = fopen(fname, "rb");
    if (!fp) {
      fprintf(stderr, " no such file '%s'\n", fname);
      exit(-1);
    }

    ivsize = metamax * (1 + reactionmax);
    iv = (int *) malloc(ivsize * sizeof(int));
    wrote = found = it = 0;
    fread(iv, sizeof(int), ivsize, fp);
    fclose(fp);

    // allocating . . .
    cl = (CL *) malloc(metamax * sizeof(CL));
    treealloc = 2 * reactionmax;
    tree = (TREE *) malloc(treealloc * sizeof(TREE));
    for (i = 0; i < treealloc; i++) {
      tree[i].alloc = tree[i].nkid = 0;
      tree[i].kid = (int *) NULL;
    }
    cl[0].node = (int *) malloc(metamax * sizeof(int));
    piv = iv;
    nclu = 1;
    tmp = (int *) calloc(3 * metamax, sizeof(int));
    ntmp = tmp + metamax;
    ctmp = ntmp + metamax;
    index = (int *) calloc(metamax, sizeof(int));

    // constructing the tree from right to left . . .

    // initialize 'tree' and 'cl' . . .
    tree[0].level = 0;
    ilevel = itree = 1;
    for (i = 1; i < metamax; i++) cl[i].size = 0;
    cl[0].size = metamax;
    xt = cl[0].tree = 0;
    for (i = 0; i < cl[0].size; i++) cl[0].node[i] = i;
    
    // one 'hi' iteration at a time . . .
    for (i = 0; i <= reactionmax; i++) {
      oldilevel = ilevel;
      // checking every cluster . . .
      event = 0;
      for (j = 0; j < nclu; j++) {
	jtmp = itmp = 0;
	for (k = 0; k < metamax; k++) tmp[k] = -1;
	for (k = 0; k < cl[j].size; k++) {
	  me = cl[j].node[k];
	  l = 0;
	  while ((tmp[l] != piv[me]) && (l < itmp)) l++;
	  // has a cluster split?
	  if (l >= itmp) {
	    ntmp[itmp] = 1;
	    tmp[itmp++] = piv[me];
	  } else ntmp[l]++;
	}
	// if a cluster has broken squeeze in the new . . .
	if (itmp > 1) {
	  event = 1;
	  inx(itmp, ntmp, index);
	  metree = cl[j].tree;
	  nctmp = cl[j].size;
	  for (k = 0; k < cl[j].size; k++) ctmp[k] = cl[j].node[k];
	  for (k = nclu - 1; k >= j + 1; k--) {
	    cl[k + (itmp - 1)].size = cl[k].size;
	    cl[k + (itmp - 1)].node = cl[k].node;
	    cl[k + (itmp - 1)].tree = cl[k].tree;
	  }
	  free(cl[j].node);
	  for (k = 0; k < itmp; k++) {
	    cl[j + k].node = (int *) malloc(ntmp[index[k]] * sizeof(int));
	    cl[j + k].size = 0;
	  }
	  for (k = 0; k < itmp; k++) {
	    for (l = 0; l < nctmp; l++) {
	      if (piv[ctmp[l]] == tmp[index[k]])
		cl[j + k].node[cl[j + k].size++] = ctmp[l];
	    }
	    // adding child to the current branching point
	    cl[j + k].tree = itree;
	    if (1 + tree[metree].nkid >= tree[metree].alloc) {
	      tree[metree].alloc += NALLOC;
	      tree[metree].kid = (int *)
		realloc(tree[metree].kid,
			tree[metree].alloc * sizeof(int));
	    }
	    tree[metree].kid[tree[metree].nkid++] = itree++;
	  }
	  jtmp += itmp - 1;
	  j += itmp - 1;

	  // checking tree's allocation . . .
	  if (itree >= treealloc) {
	    treealloc += NALLOC;
	    tree = (TREE *) realloc(tree, treealloc * sizeof(TREE));
	    for (k = treealloc - 1; k >= treealloc - NALLOC; k--) {
	      tree[k].alloc = tree[k].nkid = 0;
	      tree[k].kid = (int *) NULL;
	    }
	  }
	  nclu += jtmp;
	}
      }
      if (event) ilevel++;
      for (k = 0; k < nclu; k++) tree[cl[k].tree].level = ilevel;
      piv += metamax;
    }
    
    piv -= metamax;
    k = 0;
    for (i = 0; i < metamax; i++) for (j = 0; j < metamax; j++)
      if (piv[j] == cl[i].node[0] + 1) {
	tmp[k++] = j;
	break;
      }

    // setting the sizes . . .
    ymax = DY * metamax;
    xmax = (int) rint(ymax / RATIO);
    dx = xmax / ilevel;
    dy = ymax / metamax;

    // initialization . . .
    if (treealloc < itree + metamax) {
      tree = (TREE *) realloc(tree, (itree + metamax) * sizeof(TREE));
      for (k = treealloc - 1; k < itree + metamax; k++) {
	tree[k].alloc = tree[k].nkid = 0;
	tree[k].kid = (int *) NULL;
      }
    }

    // add leftmost nodes to the tree . . .
    for (i = 0; i < metamax; i++) {
      metree = cl[i].tree;
      if (1 + tree[metree].nkid >= tree[metree].alloc) {
	tree[metree].alloc += NALLOC;
	tree[metree].kid = (int *)
	  realloc(tree[metree].kid, tree[metree].alloc * sizeof(int));
      }
      tree[metree].kid[tree[metree].nkid++] = itree;
      tree[itree].level = ilevel;
      tree[itree++].y = dy * i;
    }

    // open and print header ti xfig file . . .
    sprintf(f2name, "fig/%s%s.fig", modetag, orgtag);
    fp = fopen(f2name, "w");
    fprintf(fp, "#FIG 3.2\nPortrait\nCenter\nMetric\n"
	    "A4\n100.00\nSingle\n-2\n1200 2\n"
	    "0 32 #c70077\n0 33 #777777\n");

    jtree = itree - metamax - 1;

    // construct the hierarchy tree graphics . . .
    for (i = itree - metamax - 1; i >= 0; i--) {
      tree[i].y = (tree[tree[i].kid[0]].y +
		   tree[tree[i].kid[tree[i].nkid - 1]].y) / 2;
      for (j = 0; j < tree[i].nkid; j++)
	if (tree[tree[i].kid[j]].level != tree[i].level)
	  fprintf(fp, LINE "%d %d %d %d\n",
		  (ilevel - tree[tree[i].kid[j]].level) * dx,
		  tree[tree[i].kid[j]].y, (ilevel - tree[i].level) * dx,
		  tree[tree[i].kid[j]].y);
      if (tree[i].level) if (tree[tree[i].kid[tree[i].nkid - 1]].y
			     != tree[tree[i].kid[0]].y)
	fprintf(fp, LINE "%d %d %d %d\n",
		(ilevel - tree[i].level) * dx,
		tree[tree[i].kid[0]].y, (ilevel - tree[i].level) * dx,
		tree[tree[i].kid[tree[i].nkid - 1]].y);
    }

    // if a file with substance names exists, print these in the picture .
    // a line number in the textfile should correspond to the index in
    // the network file . . .
    sprintf(f2name, "%s/%s.nam", modetag, orgtag);
    nfp = fopen(f2name, "r");
    if (nfp) {
      nfp = fopen(f2name, "r");
      for (i = 0; i < metamax; i++) {
	j = 0;
	while (j < tmp[i]) {while (getc(nfp) != '\n'); j++;}
	fgets(orgnames, 200, nfp);
	rewind(nfp);
	for (j = 0; j < 200; j++) if (orgnames[j] == '\n') {
	  orgnames[j] = '\0';
	  break;
	}
	fprintf(fp, TAG2, dy * i + 45, tmp[i] + 1);
	fprintf(fp, TEXTTAG, dy * i + 45, orgnames);
      }
      fclose(nfp);
    }

    // free it all . . .
    free(iv); free(tmp); free(index); fclose(fp);
    for (i = 0; i < treealloc; i++) if (tree[i].nkid) free(tree[i].kid);
    for (i = 0; i < metamax; i++) if (cl[i].size) free(cl[i].node);
    for (i = 0; i < n; i++)
      if (node[i].outdegree) free(node[i].outneighbor);
    free(tree); free(cl); free(node);
  }
  return 0;
}

// For statistics about the hierarchical ordering this is the routine
// to modify. cl[x].size is the variable to use . . .
//------------------------------------------------------------------------
