// SUBNETWORK HIERARCHIES OF BIOCHEMICAL NETWORKS ========================
// by Petter Holme, April 2002 ===========================================

#include "net.h"
#include <sys/time.h>

NODE *node;
PAR par;
int batchmode, *iv;
char fname[NAMESIZE], orgtag[4];

//------------------------------------------------------------------------
// for timing . . .

double dtime () {
  struct timeval tp;
  
  gettimeofday(&tp, 0);
  return (tp.tv_sec + 1.0e-6 * tp.tv_usec);
}

//------------------------------------------------------------------------
// parsing data files and constructing metabolic networks . . .
// these files for the WIT organisms can be downloaded from
// http://www.nd.edu/~networks/database/ the syntax is as follows:
// (1.) each line represents a directed link: 'from' 'to'
// (2.) substances are enumerated 1, 2, . . .
// (3.) reaction nodes are enumerated 1000000, 1000001, . . .

void read_meta (char *modetag, char *orgtag) {
  int i, metamax, reactionmax, me, you, ti;
  char tname[16];
  FILE *fp;

  // the name of the network file . . .
  // metabolic network files should be located in the directory 'meta' in
  // the same directory as the execuable, corresponding directory for
  // whole-cell network files is 'cell'

  // checking if the file exists (as listed in 'orgnames') . . .
  sprintf(tname, "%s/%s.dat", modetag, orgtag);
  fp = fopen(tname, "r");
  if (!fp) {
    fprintf(stderr, " couldn't open %s\n", tname);
    exit(-1);
  }

  // scanning the dimensions . . .
  // . . . number of nodes & links
  metamax = reactionmax = -1;
  while (fscanf(fp, "%d %d", &me, &you) == 2) {
    if (me < you) {
      if (you > reactionmax) reactionmax = you;
      if (me > metamax) metamax = me;
    } else {
      if (me > reactionmax) reactionmax = me;
      if (you > metamax) metamax = you;
    }
  }
  rewind(fp);
  reactionmax -= (M0 - 1);
  par.n = metamax + reactionmax;
  par.nreaction = reactionmax;
  node = (NODE *) malloc(par.n * sizeof(NODE));
  for (i = 0; i < par.n; i++) node[i].indegree = node[i].outdegree = 0;

  // . . . degrees
  ti = M0 - metamax; // moving reaction nodes to their position in 'node'
  while (fscanf(fp, "%d %d", &me, &you) == 2) {
    if (me < you) {
      node[me - 1].outdegree++;
      node[you - ti].indegree++;
    } else {
      node[me - ti].outdegree++;
      node[you - 1].indegree++;
    }
  }
  rewind(fp);

  // allocating & re-zeroing . . .
  for (i = 0; i < par.n; i++) {
    if (node[i].outdegree) node[i].outneighbor =
	    (int *) malloc(node[i].outdegree * sizeof(int));
    if (node[i].indegree) node[i].inneighbor =
	    (int *) malloc(node[i].indegree * sizeof(int));
    node[i].indegree = node[i].outdegree = 0;
  }
  
  // constructing the network . . .
  while (fscanf(fp, "%d %d", &me, &you) == 2) {
    if (me < you) {
      node[me - 1].outneighbor[node[me - 1].outdegree++] = you - ti;
      node[you - ti].inneighbor[node[you - ti].indegree++] = me - 1;
    } else {
      node[me - ti].outneighbor[node[me - ti].outdegree++] = you - 1;
      node[you - 1].inneighbor[node[you - 1].indegree++] = me - ti;
    }
  }
  fclose(fp);
}

//------------------------------------------------------------------------
// running the program . . .

void run() {
  int i = 0, *miv = (int *) NULL, strange = 0;
  double time1;
  char modetag[5];
  FILE *fp;
  
  // get organism name from the file 'orgnames' . . .
  fp = fopen("orgnames", "r");
  do {
    i++;
    if (!fgets(orgtag, 4, fp)) strange = 1;
    if ((!isalpha(orgtag[0])) || (!isalpha(orgtag[1]))) strange = 2;
  } while ((!strange) && (i != par.org));
  orgtag[2] = '\0';
  fclose(fp);
  if (strange) {
    fprintf(stderr, " organism-index out of range [1,%d] "
	    "(or error in 'orgnames')\n", i - 1);
    exit(-1);
  }
  
  switch (par.mode) {
  case META: strcpy(modetag, "meta"); break;
  case CELL: strcpy(modetag, "cell");  break;
  }
  sprintf(fname, "%s%s", modetag, orgtag);
 
  if (batchmode) printf(" datafile: '%s'", fname);
  fflush(stdout);
    
  // further initializing . . .
  node = (NODE *) NULL;

  // timing . . .
  time1 = dtime();
  
  // read network . . .
  read_meta(modetag, orgtag);

  // allocate vectors for network statistics . . .
  miv = iv = (int *) malloc((par.n - par.nreaction) *
			    (1 + par.nreaction) * sizeof(int));
  
  // measure statistics . . .
  measure();

  // save to binary file . . .
  save_data(miv);

  // how fast?
  printf(" . . . time %gs\n", dtime() - time1);
  
  // cleaning . . .
  free(node); free(miv);
}

//------------------------------------------------------------------------
// for the prompt . . .

char *gl (FILE *stream) {
#define LEN 80
  static char inpstr[LEN];

  if (stream == stdin) printf("hi> ");
  
  return fgets(inpstr, LEN, stream);
#undef LEN
}

//------------------------------------------------------------------------
// The program can be ran either interactively (writing commands at a
// prompt or with a input file, in 'batchmode'). The syntax is the same
// for both cases: 'cell 1' means that the whole-cell network should be
// calculated for organism no. 1 (of the file 'orgnames'). 'meta 2'
// means that the metabolic network for organism no. 2 should be
// calculated. '-' starts the program. An example of an input file
// follows. This input file should be located in a directory 'inp/' in
// the same directory as the executable. If the program is started by
// 'hi x' on the command line it reads the input file 'inp/x'.

/* example of input file . . .
cell 1
-
meta 2
-
*/

int main(int argc, char *argv[]) {
  FILE *stream;
  char *f1name, *s;
  
  batchmode = (argc > 1);
  
  par.org = 1;
  par.mode = META;

  if (batchmode) {
    f1name = malloc(5 + strlen(argv[1]));
    strcpy(f1name, "inp/");
    strcat(f1name, argv[1]);
    stream = fopen(f1name, "r");
    if (!stream) {
      fprintf(stderr, " can't open file: %s\n", f1name);
      return 0;
    } 
  } else stream = stdin;
  
  while (s = gl(stream)) {
    if (strncmp(s, "meta ", 5) == 0) {
      sscanf(s + 5, "%d", &par.org);
      par.mode = META;
    }
    if (strncmp(s, "cell ", 5) == 0) {
      sscanf(s + 5, "%d", &par.org);
      par.mode = CELL;
    }
    if (strncmp(s, "-", 1) == 0) run();
  } 
  if (stream != stdin) fclose(stream);
  return 0;
}

//------------------------------------------------------------------------
