// SUBNETWORK HIERARCHIES OF BIOCHEMICAL NETWORKS ========================
// by Petter Holme, April 2002 ===========================================

// this program constructs the subnetwork including all nodes adjacent
// to the reaction nodes adjacent to the input nodes; i.e. the second
// neighborhood of the input node. The output of the program is in
// Pajek's '.net' format (http://vlado.fmf.uni-lj.si/pub/networks/pajek/)

#include "net.h"
#define NALLOC 10

typedef struct PAR_P2P {
  int n, org, nets, nreaction, nclu, other1, mode;
} PAR_P2P;

typedef struct NODE_P2P {
  int indegree, outdegree, *inneighbor, *outneighbor;
} NODE_P2P;

typedef struct CLUSTER {
  int outdegree, outalloc, *outneighbor, node;
} CLUSTER;

NODE_P2P *node;
CLUSTER *cluster, *cl;
PAR_P2P par;
int batchmode, metamax, ipajek, *pajek;
char fname[50], orgtag[3];

//------------------------------------------------------------------------
// parsing data file and constructing the network . . .

void read_net (char *modetag, char *orgtag) {
  int i, reactionmax, me, you, ti;
  char tname[16];
  FILE *fp;

  sprintf(tname, "%s/%s.dat", modetag, orgtag);
  fp = fopen(tname, "r");
  if (!fp) {
    fprintf(stderr, " error opening %s\n", tname);
    exit(-1);
  }

  // scanning the dimensions . . .
  // . . . number of nodes
  metamax = reactionmax = -1;
  while (fscanf(fp, "%d %d", &me, &you) == 2) {
    if (me < you) {
      if (you > reactionmax) reactionmax = you;
      if (me > metamax) metamax = me;
    } else {
      if (me > reactionmax) reactionmax = me;
      if (you > metamax) metamax = you;
    }
  }
  rewind(fp);
  reactionmax -= (M0 - 1);
  par.n = metamax + reactionmax;
  par.nreaction = reactionmax;
  node = (NODE_P2P *) malloc(par.n * sizeof(NODE_P2P));
  for (i = 0; i < par.n; i++) node[i].indegree = node[i].outdegree = 0;

  // . . . degrees
  ti = M0 - metamax; // moving reaction nodes to their position in 'node'
  while (fscanf(fp, "%d %d", &me, &you) == 2) {
    if (me < you) {
      node[me - 1].outdegree++;
      node[you - ti].indegree++;
    } else {
      node[me - ti].outdegree++;
      node[you - 1].indegree++;
    }
  }
  rewind(fp);

  // allocating & re-zeroing . . .
  for (i = 0; i < par.n; i++) {
    if (node[i].outdegree) node[i].outneighbor =
	    (int *) malloc(node[i].outdegree * sizeof(int));
    if (node[i].indegree) node[i].inneighbor =
	    (int *) malloc(node[i].indegree * sizeof(int));
    node[i].indegree = node[i].outdegree = 0;
  }
  
  // constructing the network . . .
  while (fscanf(fp, "%d %d", &me, &you) == 2) {
    if (me < you) {
      node[me - 1].outneighbor[node[me - 1].outdegree++] = you - ti;
      node[you - ti].inneighbor[node[you - ti].indegree++] = me - 1;
    } else {
      node[me - ti].outneighbor[node[me - ti].outdegree++] = you - 1;
      node[you - 1].inneighbor[node[you - 1].indegree++] = me - ti;
    }
  }
  fclose(fp);
}

//------------------------------------------------------------------------
// saving the cluster in Pajek's format . . .
// http://vlado.fmf.uni-lj.si/pub/networks/pajek/

void write_pajek (char *modetag, char *orgtag) {
  int i, j, k;
  char orgnames[200], pname[50], f2name[50];
  FILE *pfp, *nfp;

  // opening text-file, a line number in the textfile should correspond
  // to the index in the network file . . .
  sprintf(f2name, "%s/%s.nam", modetag, orgtag);
  sprintf(pname, "pajek/%s%s.net", orgtag, modetag);
  pfp = fopen(pname, "w");

  if (!pfp) {fprintf(stderr, "no namefile\n"); exit(-1);}
  
  fprintf(pfp, "*Vertices %d\n", par.nclu);
  for (k = 0; k < ipajek; k++) {
    nfp = fopen(f2name, "r");
    if (!nfp) {
      fprintf(stderr, " no file '%s'\n", f2name);
      exit(-1);
    }
    for (i = 0; i < metamax; i++) {
      j = 0;
      while (j < pajek[k] - 1) {while (getc(nfp) != '\n'); j++;}
      fgets(orgnames, 200, nfp);
      rewind(nfp);
      for (j = 0; j < 200; j++) if (orgnames[j] == '\n') {
	orgnames[j] = '\0';
	break;
      }
    }
    fclose(nfp);
    fprintf(pfp, "%d \"%s\" ic Fuchsia\n", k + 1, orgnames);
  }
  for (i = ipajek; i < par.other1; i++)
    fprintf(pfp, "%d \"i%d\" ic White\n", i + 1, i + 1 - ipajek);
  for (k = par.other1; k < par.nclu; k++) {
    nfp = fopen(f2name, "r");
    for (i = 0; i < metamax; i++) {
      j = 0;
      while (j < cluster[k].node) {while (getc(nfp) != '\n'); j++;}
      fgets(orgnames, 200, nfp);
      rewind(nfp);
      for (j = 0; j < 200; j++) if (orgnames[j] == '\n') {
	orgnames[j] = '\0';
	break;
      }
    }
    fprintf(pfp, "%d \"%s\" ic Gray\n", k + 1, orgnames);
    fclose(nfp);
  }
  
  fprintf(pfp, "*Arcslist\n");
  for (i = 0; i < par.nclu; i++) {
    if (cluster[i].outdegree) {
      fprintf(pfp, "%d", i + 1);
      for (j = 0; j < cluster[i].outdegree; j++)
	fprintf(pfp, " %d", cluster[i].outneighbor[j] + 1);
      fprintf(pfp, "\n");
    }
  }
  fclose(pfp);
  printf("wrote cluster to '%s'\n", pname);
}

//------------------------------------------------------------------------
// the nodes of 'cluster' are ordered as follows:
// 0 -- ipajek-1 are the original nodes specified at input
// ipajek -- par.other1 - 1 are the reaction nodes
// par.other1 -- are substance nodes adjacent to the reaction nodes above
//               but not among those specified at input

void build_cluster_with_reaction_nodes () {
  int i, j, k, me, you, iyou, *reactionlist, neq, nint = 0, nother = 0;

  reactionlist = (int *) malloc(par.nreaction * sizeof(int));
  
  // find the reaction nodes adjacent to metabolites in the cluster . . .
  // looping over members of the cluster (as specified in input) . . .
  for (i = 0; i < ipajek; i++) {
    me = pajek[i] - 1;
      // add new reaction nodes in 'outneighbor'
    for (j = 0; j < node[me].outdegree; j++) {
      neq = 1;
      for (k = 0; neq && (k < nint); k++)
	neq = (node[me].outneighbor[j] != reactionlist[k]);
      if (neq) reactionlist[nint++] = node[me].outneighbor[j];
    }
    
    // add new reaction nodes in 'inneighbor'
    for (j = 0; j < node[me].indegree; j++) {
      neq = 1;
      for (k = 0; neq && (k < nint); k++)
	neq = (node[me].inneighbor[j] != reactionlist[k]);
      if (neq) reactionlist[nint++] = node[me].inneighbor[j];
    }
  }

  // indices for metabolites not in the original network . . .
  par.other1 = nint + ipajek;
  cluster = (CLUSTER *) malloc(par.n * sizeof(CLUSTER));

  // first take care of the nodes specified at input . . .
  // (note that neighbors of these will belong to the cluster)
  for (i = 0; i < ipajek; i++) {
    cluster[i].node = pajek[i] - 1;
    cluster[i].outdegree = node[pajek[i] - 1].outdegree;
    cluster[i].outneighbor =
      (int *) malloc(cluster[i].outdegree * sizeof(int));
    for (j = 0; j < cluster[i].outdegree; j++) {
      k = 0;
      while (node[pajek[i] - 1].outneighbor[j] != reactionlist[k]) k++;
      cluster[i].outneighbor[j] = k + ipajek;
    }
  }

  // then the reaction nodes . . .
  for (i = 0; i < nint; i++) {
    me = i + ipajek;
    cluster[me].node = reactionlist[i];
    cluster[me].outdegree = 0;
    cluster[me].outalloc = node[reactionlist[i]].outdegree;
    cluster[me].outneighbor =
      (int *) malloc(cluster[me].outalloc * sizeof(int));

    // looking outwards . . .
    for (j = 0; j < node[reactionlist[i]].outdegree; j++) {
      you = node[reactionlist[i]].outneighbor[j];
      neq = 1;
      for (k = 0; neq && (k < ipajek); k++)
	neq = (pajek[k] != you + 1);
      if (neq) { // the neighbor belongs to the 'other'
	// new in 'other'? . . .
	neq = 1;
	for (k = 0; neq && (k < nother); k++)
	  neq = (cluster[par.other1 + k].node != you);
	if (neq) { // . . . yes, new in 'other'
	  cluster[par.other1 + nother].outalloc = 
	    cluster[par.other1 + nother].outdegree = 0;
	  cluster[par.other1 + nother].outneighbor = (int *) NULL;
	  iyou = par.other1 + nother;
	  cluster[iyou].node = node[reactionlist[i]].outneighbor[j];
	  nother++;
	} else iyou = par.other1 + k - 1; // . . . no, present in 'other'
	cluster[me].outneighbor[cluster[me].outdegree++] = iyou;
      } else // the neighbor belongs to the input nodes
	cluster[me].outneighbor[cluster[me].outdegree++] = k - 1;
    }

    // looking inwards . . .
    for (j = 0; j < node[reactionlist[i]].indegree; j++) {
      you = node[reactionlist[i]].inneighbor[j];
      neq = 1;
      for (k = 0; neq && (k < ipajek); k++)
	neq = (pajek[k] != you + 1);
      if (neq) { // the neighbor belongs to the 'other'
	// new in 'other'?
	neq = 1;
	for (k = 0; neq && (k < nother); k++)
	  neq = (cluster[par.other1 + k].node != you);
	if (neq) { // yes, new in 'other'
	  cluster[par.other1 + nother].outalloc = 
	    cluster[par.other1 + nother].outdegree = 0;
	  cluster[par.other1 + nother].outneighbor = (int *) NULL;
	  iyou = par.other1 + nother;
	  cluster[iyou].node = you;
	  nother++;
	} else iyou = par.other1 + k - 1; // no, present in 'other'
	// allocation check . . .
	if (cluster[iyou].outalloc <= cluster[iyou].outdegree + 1) {
	  cluster[iyou].outalloc += NALLOC;
	  cluster[iyou].outneighbor =
	    (int *) realloc(cluster[iyou].outneighbor,
			    cluster[iyou].outalloc * sizeof(int));
	}
	// adding . . .
	cluster[iyou].outneighbor[cluster[iyou].outdegree++] = i + ipajek;
      }
    }
  }
  par.nclu = nother + par.other1;
  free(reactionlist);
}

//------------------------------------------------------------------------
// finds the index in 'cluster' of 'from' . . .

INLINE int my_index (int from) {
  int i, neq, cme;
  
  neq = 1;
  for (i = 0; neq && (i < ipajek); i++) neq = (cluster[i].node != from);
  if (neq) {
    cme = i - 1;
  } else {
    neq = 1;
    for (i = ipajek; neq && (i < par.other1); i++)
      neq = (cluster[i].node != from);
    if (neq) fprintf(stderr, "hmm\n");  // DBG
    cme = i - 1 + ipajek;
    if (cluster[i - 1].node != from)
      fprintf(stderr, "hmm %d %d\n", cluster[i - 1].node, from);  // DBG
  }
  return cme;
}

//------------------------------------------------------------------------

void delete_reaction_nodes () {
  int i, j, k, ime, neq, me, from, cfrom, to, cto;
  
  cl = (CLUSTER *) malloc((par.nclu - ipajek) * sizeof(CLUSTER));
  for (i = 0; i < par.nclu - ipajek; i++) {
    cl[i].node = cl[i].outdegree = cl[i].outalloc = 0;
    cl[i].outneighbor = (int *) NULL;
  }

  // add direct links (over the reaction nodes)
  for (ime = ipajek; ime < par.other1; ime++) {
    me = cluster[ime].node;
    for (i = 0; i < node[me].indegree; i++) {
      from = node[me].inneighbor[i];
      cfrom = my_index(from);
      for (j = 0; j < node[me].outdegree; j++){
	to = node[me].outneighbor[j];
	cto = my_index(to);
	neq = 1;
	for (k = 0; neq && (k < cl[cfrom].outdegree); k++)
	  neq = (cl[cfrom].outneighbor[k] != cto);
	if (neq) {
	  // allocation check . . .
	  if (cl[cfrom].outalloc <= cl[cfrom].outdegree + 1) {
	    cl[cfrom].outalloc += NALLOC;
	    cl[cfrom].outneighbor =
	      (int *) malloc(cl[cfrom].outalloc * sizeof(int));
	  }
	  // adding . . .
	  cl[cfrom].outneighbor[cl[cfrom].outdegree++] = cto;
	}
      }
    }
  }

  // move 'cl' to 'cluster'
  for (i = 0; i < par.nclu; i++) {
    if (cluster[i].outdegree) free(cluster[i].outneighbor);
  }
  for (i = 0; i < par.other1; i++) {
    cluster[i].node = cl[i].node;
    cluster[i].outdegree = cl[i].outdegree;
    cluster[i].outneighbor = cl[i].outneighbor;
    cluster[i].node = cl[i].outdegree = 0;
  }
  par.nclu -= par.nreaction;
  par.nreaction = 0;
  par.other1 = ipajek;
  free(cl);
}

//------------------------------------------------------------------------

void run () {
  int i = 0, strange = 0;
  char modetag[5];
  FILE *fp;
  
  // get organism name from the file 'orgnames' . . .
  fp = fopen("orgnames", "r");
  do {
    i++;
    if (!fgets(orgtag, 4, fp)) strange = 1;
    if ((!isalpha(orgtag[0])) || (!isalpha(orgtag[1]))) strange = 2;
  } while ((!strange) && (i != par.org));
  orgtag[2] = '\0';
  fclose(fp);
  if (strange) {
    fprintf(stderr, " organism-index out of range [1,%d] "
	    "(or error in 'orgnames')\n", i - 1);
    exit(-1);
  }
  switch (par.mode) {
  case META: sprintf(modetag, "meta"); break;
  case CELL: sprintf(modetag, "cell"); break;
  }
    
  // further initializing . . .
  node = (NODE_P2P *) NULL;
  
  // read network . . .
  read_net(modetag, orgtag);
  
  // re-build cluster . . .
  build_cluster_with_reaction_nodes();

  // write to pajek file . . .
  write_pajek(modetag, orgtag);
  
  // cleaning . . .
  for (i = 0; i < par.n; i++) {
    if (node[i].indegree) free(node[i].inneighbor);
    if (node[i].outdegree) free(node[i].outneighbor);
  }
  for (i = 0; i < par.nclu; i++)
    if (cluster[i].outdegree) free(cluster[i].outneighbor);
  free(node); free(cluster);
}

//------------------------------------------------------------------------

char *gl (FILE *stream) {
#define LEN 80
  static char inpstr[LEN];

  if (stream == stdin) printf("p2p> ");
  
  return fgets(inpstr, LEN, stream);
#undef LEN
}

//------------------------------------------------------------------------
// The program can be run interactively (no argument) or with an input
// file. The input file should be located in a directory 'inp/' in the
// same directory as the program itself. The syntax is the same whether
// the program is run interactively or not:
// 1. if '-' is typed the program starts
// 2. if an input row starts with 'meta ' the next string (should be a
//    number) gives the line number in an indexfile 'orgnames' (in the
//    directory of the program) of organism names
//    if an input row starts with 'cell ' the same thing is done for a
//    whole-cell network
// 3. any other line of input is a number of a node included in the
//    cluster

/* example of input file . . .
meta 2
12
13
101
-
*/

int main (int argc, char *argv[]) {
  int palloc;
  FILE *stream; char *f1name, *s;
  
  batchmode = (argc > 1);
  ipajek = palloc = 0;
  pajek = (int *) NULL;
  par.org = 1;
  par.mode = META;
  
  if (batchmode) {
    f1name = malloc(5 + strlen(argv[1]));
    strcpy(f1name, "inp/");
    strcat(f1name, argv[1]);
    stream = fopen(f1name, "r");
    if (!stream) {
      fprintf(stderr, " can't open file: %s\n", f1name);
      return 0;
    } 
  } else stream = stdin;
  
  while (s = gl(stream)) {
    if (strncmp(s, "meta ", 5) == 0) {
      par.mode = META;
      sscanf(s + 5, "%d", &par.org);
    } else {
      if (strncmp(s, "cell ", 5) == 0) {
	par.mode = CELL;
	sscanf(s + 5, "%d", &par.org);
      } else {
	if (strncmp(s, "-", 1) == 0) run();
	else {
	  if (1 + ipajek >= palloc) {
	    palloc += NALLOC;
	    pajek = (int *) realloc(pajek, palloc * sizeof(int));
	  }
	  sscanf(s, "%d", &pajek[ipajek++]);
	}
      }
    }
  }
  if (stream != stdin) fclose(stream);
  return 0;
}

//------------------------------------------------------------------------
